
// Type-secure C++ interface for multi-dimensional arrays.

extern "C" {
  #include "armci.h"
}

void malloc_rm(void* ptrs [], MPI_Comm *apg_comm, const int bytes) ;
void free_rm(void* ptrs [], MPI_Comm *apg_comm) ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// template classes `SectionRMR': Add subscripting operations to       //
//                                `SectionRM_'.                        //
//                                                                     //
// template classes `ArrayRMR': Add element allocation to `SectionRMR' //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
class SectionRM0 : public Section0<T> {
public :
  // Constructors

  SectionRM0(const SectionRM0<T>& a) : Section0<T>(a),
                                       offset(a.offset), ptrs(a.ptrs) {}

  SectionRM0(const Group& p,
             T* _data,
             const int _offset, void** _ptrs,
             const int disp) : Section0<T>(p,
                                           _data + disp),
                               offset(_offset + disp), ptrs(_ptrs) {}

  // Simple subscripting

  T& operator()() const {
    return dat() [0] ;
  }

  // Section subscripting

  SectionRM0<T> sect() const {
    return SectionRM0<T>(grp(),
                         dat(),
                         offset, ptrs,
                         0) ;
  }

protected :
  SectionRM0() {}  // Used by subclasses.

  SectionRM0(const Group& p) : Section0<T>(p) {}
      // Used by Array classes to initialize virtual base.

public :
  int offset ;
  void** ptrs ;
} ;


template<class T>
class ArrayRM0 : public SectionRM0<T> {
public:

  ArrayRM0(const Group& p = apg) : SectionRM0<T>(p) {

    int size = 1 ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM0() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM0(const ArrayRM0& a) : SectionRM0(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM1 : public Section1<T> {
public :
  // Constructors

  SectionRM1(const SectionRM1<T>& a) : Section1<T>(a),
                                       offset(a.offset), ptrs(a.ptrs) {}

  SectionRM1(const Range& x_0, const Group& p,
             const int str_0, T* _data,
             const int _offset, void** _ptrs,
             const int disp) : Section1<T>(x_0, p,
                                           str_0, _data + disp),
                               offset(_offset + disp), ptrs(_ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0) const {
    return dat() [map(0).offset(i_0)] ;
  }

  // Section subscripting

  SectionRM1<T> sect(const Trip& i_0) const {
    return SectionRM1<T>(rng(0)(i_0), grp(),
                         str(0), dat(),
                         offset, ptrs,
                         0) ;
  }

  SectionRM0<T> sect(const Location& i_0) const {
    return SectionRM0<T>(grp() / i_0,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0)) ;
  }

protected :
  SectionRM1() {}  // Used by subclasses.

  SectionRM1(const Group& p) : Section1<T>(p) {}
      // Used by Array classes to initialize virtual base.

public :
  int offset ;
  void** ptrs ;
} ;


template<class T>
class ArrayRM1 : public SectionRM1<T> {
public:

  ArrayRM1(const Range& x_0, const Group& p = apg) : SectionRM1<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM1() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM1(const ArrayRM1& a) : SectionRM1(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM1S : public virtual Section1<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM1S(const SectionRM1S<T>& a) : SectionRM1<T>(a) {}

  SectionRM1S(const Range& x_0, const Group& p,
              const int str_0, T* _data,
              const int offset, void** ptrs,
              const int disp) : SectionRM1<T>(x_0, p,
                                              str_0, _data + disp,
                                              offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0) const {
    return dat() [map(0).offset(rng(0)(i_0))] ;
  }

  // Section subscripting

  SectionRM1S<T> sect(const Trip& i_0) const {
    return SectionRM1S<T>(rng(0)(i_0), grp(),
                          str(0), dat(),
                          offset, ptrs,
                          0) ;
  }

  SectionRM0<T> sect(const int i_0) const {
    return SectionRM0<T>(grp(),
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0))) ;
  }

protected :
  SectionRM1S() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM1S : public SectionRM1S<T>, private virtual SectionRM1<T> {
public:

  ArrayRM1S(const int n_0, const Group& p = apg) : SectionRM1<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM1S() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM1S(const ArrayRM1S& a) : SectionRM1S(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM2 : public Section2<T> {
public :
  // Constructors

  SectionRM2(const SectionRM2<T>& a) : Section2<T>(a),
                                       offset(a.offset), ptrs(a.ptrs) {}

  SectionRM2(const Range& x_0,
             const Range& x_1, const Group& p,
             const int str_0,
             const int str_1, T* _data,
             const int _offset, void** _ptrs,
             const int disp) : Section2<T>(x_0,
                                           x_1, p,
                                           str_0,
                                           str_1, _data + disp),
                               offset(_offset + disp), ptrs(_ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(i_1)] ;
  }

  // Section subscripting

  SectionRM2<T> sect(const Trip& i_0,
                     const Trip& i_1) const {
    return SectionRM2<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         str(0),
                         str(1), dat(),
                         offset, ptrs,
                         0) ;
  }

  SectionRM1<T> sect(const Location& i_0,
                     const Trip& i_1) const {
    return SectionRM1<T>(rng(1)(i_1), grp() / i_0,
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(i_0)) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const Location& i_1) const {
    return SectionRM1<T>(rng(0)(i_0), grp() / i_1,
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(i_1)) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const Location& i_1) const {
    return SectionRM0<T>(grp() / i_0 / i_1,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(i_1)) ;
  }

protected :
  SectionRM2() {}  // Used by subclasses.

  SectionRM2(const Group& p) : Section2<T>(p) {}
      // Used by Array classes to initialize virtual base.

public :
  int offset ;
  void** ptrs ;
} ;


template<class T>
class ArrayRM2 : public SectionRM2<T> {
public:

  ArrayRM2(const Range& x_0,
           const Range& x_1, const Group& p = apg) : SectionRM2<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM2() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM2(const ArrayRM2& a) : SectionRM2(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM2SD : public virtual Section2<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM2SD(const SectionRM2SD<T>& a) : SectionRM2<T>(a) {}

  SectionRM2SD(const Range& x_0,
               const Range& x_1, const Group& p,
               const int str_0,
               const int str_1, T* _data,
               const int offset, void** ptrs,
               const int disp) : SectionRM2<T>(x_0,
                                               x_1, p,
                                               str_0,
                                               str_1, _data + disp,
                                               offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(i_1)] ;
  }

  // Section subscripting

  SectionRM2SD<T> sect(const Trip& i_0,
                       const Trip& i_1) const {
    return SectionRM2SD<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           0) ;
  }

  SectionRM1<T> sect(const int i_0,
                     const Trip& i_1) const {
    return SectionRM1<T>(rng(1)(i_1), grp(),
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const Location& i_1) const {
    return SectionRM1S<T>(rng(0)(i_0), grp() / i_1,
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(i_1)) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const Location& i_1) const {
    return SectionRM0<T>(grp() / i_1,
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(i_1)) ;
  }

protected :
  SectionRM2SD() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM2SD : public SectionRM2SD<T>, private virtual SectionRM2<T> {
public:

  ArrayRM2SD(const int n_0,
             const Range& x_1, const Group& p = apg) : SectionRM2<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM2SD() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM2SD(const ArrayRM2SD& a) : SectionRM2SD(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM2DS : public virtual Section2<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM2DS(const SectionRM2DS<T>& a) : SectionRM2<T>(a) {}

  SectionRM2DS(const Range& x_0,
               const Range& x_1, const Group& p,
               const int str_0,
               const int str_1, T* _data,
               const int offset, void** ptrs,
               const int disp) : SectionRM2<T>(x_0,
                                               x_1, p,
                                               str_0,
                                               str_1, _data + disp,
                                               offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(rng(1)(i_1))] ;
  }

  // Section subscripting

  SectionRM2DS<T> sect(const Trip& i_0,
                       const Trip& i_1) const {
    return SectionRM2DS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           0) ;
  }

  SectionRM1S<T> sect(const Location& i_0,
                      const Trip& i_1) const {
    return SectionRM1S<T>(rng(1)(i_1), grp() / i_0,
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(i_0)) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const int i_1) const {
    return SectionRM1<T>(rng(0)(i_0), grp(),
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const int i_1) const {
    return SectionRM0<T>(grp() / i_0,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(rng(1)(i_1))) ;
  }

protected :
  SectionRM2DS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM2DS : public SectionRM2DS<T>, private virtual SectionRM2<T> {
public:

  ArrayRM2DS(const Range& x_0,
             const int n_1, const Group& p = apg) : SectionRM2<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM2DS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM2DS(const ArrayRM2DS& a) : SectionRM2DS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM2SS : private virtual Section2<T>,
                     public Section2DS<T>,
                     public Section2SD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM2SS(const SectionRM2SS<T>& a) : SectionRM2<T>(a) {}

  SectionRM2SS(const Range& x_0,
               const Range& x_1, const Group& p,
               const int str_0,
               const int str_1, T* _data,
               const int offset, void** ptrs,
               const int disp) : SectionRM2<T>(x_0,
                                               x_1, p,
                                               str_0,
                                               str_1, _data + disp,
                                               offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(rng(1)(i_1))] ;
  }

  // Section subscripting

  SectionRM2SS<T> sect(const Trip& i_0,
                       const Trip& i_1) const {
    return SectionRM2SS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           0) ;
  }

  SectionRM1S<T> sect(const int i_0,
                      const Trip& i_1) const {
    return SectionRM1S<T>(rng(1)(i_1), grp(),
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const int i_1) const {
    return SectionRM1S<T>(rng(0)(i_0), grp(),
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const int i_1) const {
    return SectionRM0<T>(grp(),
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(rng(1)(i_1))) ;
  }

protected :
  SectionRM2SS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM2SS : public SectionRM2SS<T>, private virtual SectionRM2<T> {
public:

  ArrayRM2SS(const int n_0,
             const int n_1, const Group& p = apg) : SectionRM2<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM2SS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM2SS(const ArrayRM2SS& a) : SectionRM2SS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3 : public Section3<T> {
public :
  // Constructors

  SectionRM3(const SectionRM3<T>& a) : Section3<T>(a),
                                       offset(a.offset), ptrs(a.ptrs) {}

  SectionRM3(const Range& x_0,
             const Range& x_1,
             const Range& x_2, const Group& p,
             const int str_0,
             const int str_1,
             const int str_2, T* _data,
             const int _offset, void** _ptrs,
             const int disp) : Section3<T>(x_0,
                                           x_1,
                                           x_2, p,
                                           str_0,
                                           str_1,
                                           str_2, _data + disp),
                               offset(_offset + disp), ptrs(_ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1,
                const Location& i_2) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(i_1) +
                  map(2).offset(i_2)] ;
  }

  // Section subscripting

  SectionRM3<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return SectionRM3<T>(rng(0)(i_0),
                         rng(1)(i_1),
                         rng(2)(i_2), grp(),
                         str(0),
                         str(1),
                         str(2), dat(),
                         offset, ptrs,
                         0) ;
  }

  SectionRM2<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return SectionRM2<T>(rng(1)(i_1),
                         rng(2)(i_2), grp() / i_0,
                         str(1),
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(i_0)) ;
  }

  SectionRM2<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return SectionRM2<T>(rng(0)(i_0),
                         rng(2)(i_2), grp() / i_1,
                         str(0),
                         str(2), dat(),
                         offset, ptrs,
                         map(1).offset(i_1)) ;
  }

  SectionRM1<T> sect(const Location& i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return SectionRM1<T>(rng(2)(i_2), grp() / i_0 / i_1,
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(i_1)) ;
  }

  SectionRM2<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return SectionRM2<T>(rng(0)(i_0),
                         rng(1)(i_1), grp() / i_2,
                         str(0),
                         str(1), dat(),
                         offset, ptrs,
                         map(2).offset(i_2)) ;
  }

  SectionRM1<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return SectionRM1<T>(rng(1)(i_1), grp() / i_0 / i_2,
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(2).offset(i_2)) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const Location& i_2) const {
    return SectionRM1<T>(rng(0)(i_0), grp() / i_1 / i_2,
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(i_1) +
                         map(2).offset(i_2)) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const Location& i_1,
                     const Location& i_2) const {
    return SectionRM0<T>(grp() / i_0 / i_1 / i_2,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(i_1) +
                         map(2).offset(i_2)) ;
  }

protected :
  SectionRM3() {}  // Used by subclasses.

  SectionRM3(const Group& p) : Section3<T>(p) {}
      // Used by Array classes to initialize virtual base.

public :
  int offset ;
  void** ptrs ;
} ;


template<class T>
class ArrayRM3 : public SectionRM3<T> {
public:

  ArrayRM3(const Range& x_0,
           const Range& x_1,
           const Range& x_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3(const ArrayRM3& a) : SectionRM3(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3SDD : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3SDD(const SectionRM3SDD<T>& a) : SectionRM3<T>(a) {}

  SectionRM3SDD(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1,
                const Location& i_2) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(i_1) +
                  map(2).offset(i_2)] ;
  }

  // Section subscripting

  SectionRM3SDD<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3SDD<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2<T> sect(const int i_0,
                     const Trip& i_1,
                     const Trip& i_2) const {
    return SectionRM2<T>(rng(1)(i_1),
                         rng(2)(i_2), grp(),
                         str(1),
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM2SD<T> sect(const Trip& i_0,
                       const Location& i_1,
                       const Trip& i_2) const {
    return SectionRM2SD<T>(rng(0)(i_0),
                           rng(2)(i_2), grp() / i_1,
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(i_1)) ;
  }

  SectionRM1<T> sect(const int i_0,
                     const Location& i_1,
                     const Trip& i_2) const {
    return SectionRM1<T>(rng(2)(i_2), grp() / i_1,
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(i_1)) ;
  }

  SectionRM2SD<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const Location& i_2) const {
    return SectionRM2SD<T>(rng(0)(i_0),
                           rng(1)(i_1), grp() / i_2,
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(i_2)) ;
  }

  SectionRM1<T> sect(const int i_0,
                     const Trip& i_1,
                     const Location& i_2) const {
    return SectionRM1<T>(rng(1)(i_1), grp() / i_2,
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(2).offset(i_2)) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const Location& i_1,
                      const Location& i_2) const {
    return SectionRM1S<T>(rng(0)(i_0), grp() / i_1 / i_2,
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(i_1) +
                          map(2).offset(i_2)) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const Location& i_1,
                     const Location& i_2) const {
    return SectionRM0<T>(grp() / i_1 / i_2,
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(i_1) +
                         map(2).offset(i_2)) ;
  }

protected :
  SectionRM3SDD() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3SDD : public SectionRM3SDD<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3SDD(const int n_0,
              const Range& x_1,
              const Range& x_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3SDD() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3SDD(const ArrayRM3SDD& a) : SectionRM3SDD(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3DSD : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3DSD(const SectionRM3DSD<T>& a) : SectionRM3<T>(a) {}

  SectionRM3DSD(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1,
                const Location& i_2) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(rng(1)(i_1)) +
                  map(2).offset(i_2)] ;
  }

  // Section subscripting

  SectionRM3DSD<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3DSD<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2SD<T> sect(const Location& i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2SD<T>(rng(1)(i_1),
                           rng(2)(i_2), grp() / i_0,
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(i_0)) ;
  }

  SectionRM2<T> sect(const Trip& i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return SectionRM2<T>(rng(0)(i_0),
                         rng(2)(i_2), grp(),
                         str(0),
                         str(2), dat(),
                         offset, ptrs,
                         map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM1<T> sect(const Location& i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return SectionRM1<T>(rng(2)(i_2), grp() / i_0,
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM2DS<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const Location& i_2) const {
    return SectionRM2DS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp() / i_2,
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(i_2)) ;
  }

  SectionRM1S<T> sect(const Location& i_0,
                      const Trip& i_1,
                      const Location& i_2) const {
    return SectionRM1S<T>(rng(1)(i_1), grp() / i_0 / i_2,
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(i_0) +
                          map(2).offset(i_2)) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const int i_1,
                     const Location& i_2) const {
    return SectionRM1<T>(rng(0)(i_0), grp() / i_2,
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(i_2)) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const int i_1,
                     const Location& i_2) const {
    return SectionRM0<T>(grp() / i_0 / i_2,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(i_2)) ;
  }

protected :
  SectionRM3DSD() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3DSD : public SectionRM3DSD<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3DSD(const Range& x_0,
              const int n_1,
              const Range& x_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3DSD() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3DSD(const ArrayRM3DSD& a) : SectionRM3DSD(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3SSD : private virtual Section3<T>,
                      public Section3DSD<T>,
                      public Section3SDD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3SSD(const SectionRM3SSD<T>& a) : SectionRM3<T>(a) {}

  SectionRM3SSD(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1,
                const Location& i_2) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(rng(1)(i_1)) +
                  map(2).offset(i_2)] ;
  }

  // Section subscripting

  SectionRM3SSD<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3SSD<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2SD<T> sect(const int i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2SD<T>(rng(1)(i_1),
                           rng(2)(i_2), grp(),
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM2SD<T> sect(const Trip& i_0,
                       const int i_1,
                       const Trip& i_2) const {
    return SectionRM2SD<T>(rng(0)(i_0),
                           rng(2)(i_2), grp(),
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM1<T> sect(const int i_0,
                     const int i_1,
                     const Trip& i_2) const {
    return SectionRM1<T>(rng(2)(i_2), grp(),
                         str(2), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM2SS<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const Location& i_2) const {
    return SectionRM2SS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp() / i_2,
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(i_2)) ;
  }

  SectionRM1S<T> sect(const int i_0,
                      const Trip& i_1,
                      const Location& i_2) const {
    return SectionRM1S<T>(rng(1)(i_1), grp() / i_2,
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(rng(0)(i_0)) +
                          map(2).offset(i_2)) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const int i_1,
                      const Location& i_2) const {
    return SectionRM1S<T>(rng(0)(i_0), grp() / i_2,
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(rng(1)(i_1)) +
                          map(2).offset(i_2)) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const int i_1,
                     const Location& i_2) const {
    return SectionRM0<T>(grp() / i_2,
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(i_2)) ;
  }

protected :
  SectionRM3SSD() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3SSD : public SectionRM3SSD<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3SSD(const int n_0,
              const int n_1,
              const Range& x_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3SSD() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3SSD(const ArrayRM3SSD& a) : SectionRM3SSD(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3DDS : public virtual Section3<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3DDS(const SectionRM3DDS<T>& a) : SectionRM3<T>(a) {}

  SectionRM3DDS(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const Location& i_1,
                const int i_2) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(i_1) +
                  map(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  SectionRM3DDS<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3DDS<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2DS<T> sect(const Location& i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2DS<T>(rng(1)(i_1),
                           rng(2)(i_2), grp() / i_0,
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(i_0)) ;
  }

  SectionRM2DS<T> sect(const Trip& i_0,
                       const Location& i_1,
                       const Trip& i_2) const {
    return SectionRM2DS<T>(rng(0)(i_0),
                           rng(2)(i_2), grp() / i_1,
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(i_1)) ;
  }

  SectionRM1S<T> sect(const Location& i_0,
                      const Location& i_1,
                      const Trip& i_2) const {
    return SectionRM1S<T>(rng(2)(i_2), grp() / i_0 / i_1,
                          str(2), dat(),
                          offset, ptrs,
                          map(0).offset(i_0) +
                          map(1).offset(i_1)) ;
  }

  SectionRM2<T> sect(const Trip& i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return SectionRM2<T>(rng(0)(i_0),
                         rng(1)(i_1), grp(),
                         str(0),
                         str(1), dat(),
                         offset, ptrs,
                         map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1<T> sect(const Location& i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return SectionRM1<T>(rng(1)(i_1), grp() / i_0,
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const Location& i_1,
                     const int i_2) const {
    return SectionRM1<T>(rng(0)(i_0), grp() / i_1,
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(i_1) +
                         map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const Location& i_1,
                     const int i_2) const {
    return SectionRM0<T>(grp() / i_0 / i_1,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(i_1) +
                         map(2).offset(rng(2)(i_2))) ;
  }

protected :
  SectionRM3DDS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3DDS : public SectionRM3DDS<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3DDS(const Range& x_0,
              const Range& x_1,
              const int n_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3DDS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3DDS(const ArrayRM3DDS& a) : SectionRM3DDS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3SDS : private virtual Section3<T>,
                      public Section3DDS<T>,
                      public Section3SDD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3SDS(const SectionRM3SDS<T>& a) : SectionRM3<T>(a) {}

  SectionRM3SDS(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const Location& i_1,
                const int i_2) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(i_1) +
                  map(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  SectionRM3SDS<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3SDS<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2DS<T> sect(const int i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2DS<T>(rng(1)(i_1),
                           rng(2)(i_2), grp(),
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM2SS<T> sect(const Trip& i_0,
                       const Location& i_1,
                       const Trip& i_2) const {
    return SectionRM2SS<T>(rng(0)(i_0),
                           rng(2)(i_2), grp() / i_1,
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(i_1)) ;
  }

  SectionRM1S<T> sect(const int i_0,
                      const Location& i_1,
                      const Trip& i_2) const {
    return SectionRM1S<T>(rng(2)(i_2), grp() / i_1,
                          str(2), dat(),
                          offset, ptrs,
                          map(0).offset(rng(0)(i_0)) +
                          map(1).offset(i_1)) ;
  }

  SectionRM2SD<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const int i_2) const {
    return SectionRM2SD<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1<T> sect(const int i_0,
                     const Trip& i_1,
                     const int i_2) const {
    return SectionRM1<T>(rng(1)(i_1), grp(),
                         str(1), dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const Location& i_1,
                      const int i_2) const {
    return SectionRM1S<T>(rng(0)(i_0), grp() / i_1,
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(i_1) +
                          map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const Location& i_1,
                     const int i_2) const {
    return SectionRM0<T>(grp() / i_1,
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(i_1) +
                         map(2).offset(rng(2)(i_2))) ;
  }

protected :
  SectionRM3SDS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3SDS : public SectionRM3SDS<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3SDS(const int n_0,
              const Range& x_1,
              const int n_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3SDS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3SDS(const ArrayRM3SDS& a) : SectionRM3SDS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3DSS : private virtual Section3<T>,
                      public Section3DDS<T>,
                      public Section3DSD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3DSS(const SectionRM3DSS<T>& a) : SectionRM3<T>(a) {}

  SectionRM3DSS(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const Location& i_0,
                const int i_1,
                const int i_2) const {
    return dat() [map(0).offset(i_0) +
                  map(1).offset(rng(1)(i_1)) +
                  map(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  SectionRM3DSS<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3DSS<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2SS<T> sect(const Location& i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2SS<T>(rng(1)(i_1),
                           rng(2)(i_2), grp() / i_0,
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(i_0)) ;
  }

  SectionRM2DS<T> sect(const Trip& i_0,
                       const int i_1,
                       const Trip& i_2) const {
    return SectionRM2DS<T>(rng(0)(i_0),
                           rng(2)(i_2), grp(),
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM1S<T> sect(const Location& i_0,
                      const int i_1,
                      const Trip& i_2) const {
    return SectionRM1S<T>(rng(2)(i_2), grp() / i_0,
                          str(2), dat(),
                          offset, ptrs,
                          map(0).offset(i_0) +
                          map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM2DS<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const int i_2) const {
    return SectionRM2DS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1S<T> sect(const Location& i_0,
                      const Trip& i_1,
                      const int i_2) const {
    return SectionRM1S<T>(rng(1)(i_1), grp() / i_0,
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(i_0) +
                          map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1<T> sect(const Trip& i_0,
                     const int i_1,
                     const int i_2) const {
    return SectionRM1<T>(rng(0)(i_0), grp(),
                         str(0), dat(),
                         offset, ptrs,
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM0<T> sect(const Location& i_0,
                     const int i_1,
                     const int i_2) const {
    return SectionRM0<T>(grp() / i_0,
                         dat(),
                         offset, ptrs,
                         map(0).offset(i_0) +
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(rng(2)(i_2))) ;
  }

protected :
  SectionRM3DSS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3DSS : public SectionRM3DSS<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3DSS(const Range& x_0,
              const int n_1,
              const int n_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3DSS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3DSS(const ArrayRM3DSS& a) : SectionRM3DSS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

template<class T>
class SectionRM3SSS : private virtual Section3<T>,
                      public Section3DSS<T>,
                      public Section3SDS<T>,
                      public Section3SSD<T> {
public :
  // Constructors

  // (Due to explicit call to constructor for virtual base class,
  //  following constructors are ineffective if invoked from constructor
  //  of derived-class object).

  SectionRM3SSS(const SectionRM3SSS<T>& a) : SectionRM3<T>(a) {}

  SectionRM3SSS(const Range& x_0,
                const Range& x_1,
                const Range& x_2, const Group& p,
                const int str_0,
                const int str_1,
                const int str_2, T* _data,
                const int offset, void** ptrs,
                const int disp) : SectionRM3<T>(x_0,
                                                x_1,
                                                x_2, p,
                                                str_0,
                                                str_1,
                                                str_2, _data + disp,
                                                offset + disp, ptrs) {}

  // Simple subscripting

  T& operator()(const int i_0,
                const int i_1,
                const int i_2) const {
    return dat() [map(0).offset(rng(0)(i_0)) +
                  map(1).offset(rng(1)(i_1)) +
                  map(2).offset(rng(2)(i_2))] ;
  }

  // Section subscripting

  SectionRM3SSS<T> sect(const Trip& i_0,
                        const Trip& i_1,
                        const Trip& i_2) const {
    return SectionRM3SSS<T>(rng(0)(i_0),
                            rng(1)(i_1),
                            rng(2)(i_2), grp(),
                            str(0),
                            str(1),
                            str(2), dat(),
                            offset, ptrs,
                            0) ;
  }

  SectionRM2SS<T> sect(const int i_0,
                       const Trip& i_1,
                       const Trip& i_2) const {
    return SectionRM2SS<T>(rng(1)(i_1),
                           rng(2)(i_2), grp(),
                           str(1),
                           str(2), dat(),
                           offset, ptrs,
                           map(0).offset(rng(0)(i_0))) ;
  }

  SectionRM2SS<T> sect(const Trip& i_0,
                       const int i_1,
                       const Trip& i_2) const {
    return SectionRM2SS<T>(rng(0)(i_0),
                           rng(2)(i_2), grp(),
                           str(0),
                           str(2), dat(),
                           offset, ptrs,
                           map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM1S<T> sect(const int i_0,
                      const int i_1,
                      const Trip& i_2) const {
    return SectionRM1S<T>(rng(2)(i_2), grp(),
                          str(2), dat(),
                          offset, ptrs,
                          map(0).offset(rng(0)(i_0)) +
                          map(1).offset(rng(1)(i_1))) ;
  }

  SectionRM2SS<T> sect(const Trip& i_0,
                       const Trip& i_1,
                       const int i_2) const {
    return SectionRM2SS<T>(rng(0)(i_0),
                           rng(1)(i_1), grp(),
                           str(0),
                           str(1), dat(),
                           offset, ptrs,
                           map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1S<T> sect(const int i_0,
                      const Trip& i_1,
                      const int i_2) const {
    return SectionRM1S<T>(rng(1)(i_1), grp(),
                          str(1), dat(),
                          offset, ptrs,
                          map(0).offset(rng(0)(i_0)) +
                          map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM1S<T> sect(const Trip& i_0,
                      const int i_1,
                      const int i_2) const {
    return SectionRM1S<T>(rng(0)(i_0), grp(),
                          str(0), dat(),
                          offset, ptrs,
                          map(1).offset(rng(1)(i_1)) +
                          map(2).offset(rng(2)(i_2))) ;
  }

  SectionRM0<T> sect(const int i_0,
                     const int i_1,
                     const int i_2) const {
    return SectionRM0<T>(grp(),
                         dat(),
                         offset, ptrs,
                         map(0).offset(rng(0)(i_0)) +
                         map(1).offset(rng(1)(i_1)) +
                         map(2).offset(rng(2)(i_2))) ;
  }

protected :
  SectionRM3SSS() {}  // Used by subclasses.
} ;


template<class T>
class ArrayRM3SSS : public SectionRM3SSS<T>, private virtual SectionRM3<T> {
public:

  ArrayRM3SSS(const int n_0,
              const int n_1,
              const int n_2, const Group& p = apg) : SectionRM3<T>(p) {

    int size = 1 ;
    maps [0] = Map(x_0, size) ;
    size *= x_0.volume() ;
    maps [1] = Map(x_1, size) ;
    size *= x_1.volume() ;
    maps [2] = Map(x_2, size) ;
    size *= x_2.volume() ;

    ptrs = new void* [AdlibNProcs()] ;
    malloc_rm(ptrs, &comm, sizeof(T) * size) ;
    data = (T*) ptrs [AdlibLocId()] ;

    offset = 0 ;
  }

  ~ArrayRM3SSS() {
    free_rm(ptrs, &comm) ;
    delete [] ptrs ;
  }

private :

  ArrayRM3SSS(const ArrayRM3SSS& a) : SectionRM3SSS(a) {}  // Prohibit copying.

  MPI_Comm comm ;  // Used by 'malloc_rm', 'free_rm'
} ;

